/**      Cartman2h.C by Alexander Myasnikov                        **/
/**      Special version Cartman 2.DDP.4 block cipher              **/
/**      Cartman encrypts 128 bit block with 256..512 bit key      **/
/**      Full key size is keysizebits + 128 bit operation key      **/
/**      Gives much better transforming with new key schedule      **/
/**      Includes unique key-dependent transformation rounds       **/
/**      Freeware for any use,  non-patented, opensource cipher    **/
/**      WEB:       www.alexanderwdark.narod.ru                    **/


#include <stdlib.h>
#include <string.h>

typedef unsigned long u32;
typedef unsigned short u16;
typedef unsigned char u8;

#define keysizebits     384
#define keysizebytes    keysizebits / 8
#define keysize         keysizebytes / 4
#define subkeysize      keysize / 2

#define ror32(N, R)    _lrotr(N, R)
#define rol32(N, R)    _lrotl(N, R)

_inline u8
ror8 (u8 x, u8 y)
{
  u8 result;
  _asm
  {
    mov al, x;
    mov cl, y;
    ror al, cl;
    mov result, al;
  }
  return result;
}

_inline u8
rol8 (u8 x, u8 y)
{
  u8 result;
  _asm
  {
    mov al, x;
    mov cl, y;
    rol al, cl;
    mov result, al;
  }
  return result;
}


typedef union
{
  u32 d32;
  u8 d8[4];
  u16 d16[2];
} r32;


typedef union
{
  u32 d32[4];
  u8 d8[16];
  u16 d16[8];
} keystream;

u32 idx_box1[subkeysize];
u32 idx_box2[subkeysize];
u32 idx_box3[subkeysize];
u32 idx_box4[subkeysize];
u32 rot_box2[subkeysize - 1];
u32 rot_box3[subkeysize - 1];
u32 rot_box[keysize];

union
{
  u8 d8[keysizebytes];
  u32 d32[keysize];
}
key;

u32 opkey[4];

u32 subkey1[subkeysize];
u32 subkey2[subkeysize];

u32 sbidx1[subkeysize][subkeysize];
u32 sbidx2[subkeysize][subkeysize];

u32 idx1null;
u32 idx2null;
u32 idx3null;

typedef unsigned char cartmansb[256];

u32 oprec[subkeysize][subkeysize][8];

#define add 0
#define sub 1
#define xor 2
#define mixadd 3
#define mixsub 4
#define mixxor 5
#define mix8add 6
#define mix8sub 7
#define mix8xor 8


typedef struct
{
  u32 idx1;
  u32 idx2;
  u32 idx3;
  u32 idx4;
  u32 idx5;
  u32 idx6;
  u32 idx5_1;
  u32 idx6_1;
  u32 k1;
  u32 k2;
  u32 k3;
  u32 k4;
  u32 rot1;
  u32 rot2;
  u32 rot_box2;
  u32 rot_box3;

}
idxblock;

#define rounds_total subkeysize * subkeysize

idxblock prefetch[rounds_total];

#include "SBox.h"


void makebox (void);
void makestruct (void);



void
makebox ()
{
  u32 i, n, cnt, x;
  u32 pi, idx1, idx2, idx3, idx4, idx5, idx6;

  idx1null = 1;
  idx2null = 1;
  idx3null = 0;


  for (i = 0; i != subkeysize; i++)
    {
      idx_box1[i] =
	(ror32
	 (key.d32[i] + ((key.d32[i] << 4) ^ (key.d32[i] >> 5)),
	  (i % 31) + 1)) % (subkeysize - 1);

      idx1null = (idx1null * (idx1null + 1) >> 1) + key.d32[i];
    }

  for (i = subkeysize; i != keysize; i++)
    {
      idx_box2[(keysize - 1) - i] =
	(rol32
	 (key.d32[i] + ((key.d32[i] << 4) ^ (key.d32[i] >> 5)),
	  (i % 31) + 1)) % (subkeysize - 1);
      idx2null = (idx2null * (idx2null + 1) >> 1) + key.d32[i];
    }

  for (i = 0; i != keysize; i++)
    {
      rot_box[i] =
	((rol32 (key.d32[i] ^ (i + 1), (((keysize - 1) - i) % 31) + 1)) %
	 31) + 1;
      idx3null += key.d32[i] * (i + 1);
    }

  idx1null = (idx1null ^ (idx1null >> 16)) % subkeysize;
  idx2null = (idx2null ^ (idx2null >> 16)) % subkeysize;
  idx3null = (idx3null ^ (idx3null >> 16)) % subkeysize;

  for (i = 0; i < subkeysize - 1; i++)
    {
      rot_box2[i] = (subkey2[(subkeysize - 1) - i] % 31) + 1;
      rot_box3[i] = (subkey1[(subkeysize - 1) - i] % 31) + 1;
    }


  for (i = 0; i != subkeysize; i++)
    {
      idx_box3[i] =
	(rol32
	 (key.d32[i] + ((key.d32[i] << 5) ^ (key.d32[i] >> 9)),
	  (i % 31) + 1)) % (subkeysize - 1);
    }

  for (i = subkeysize; i != keysize; i++)
    {
      idx_box4[(keysize - 1) - i] =
	(ror32
	 (key.d32[i] + ((key.d32[i] << 5) ^ (key.d32[i] >> 9)),
	  (i % 31) + 1)) % (subkeysize - 1);
    }



  cnt = 0;

  for (i = 0; i != subkeysize; i++)
    for (n = 0; n != subkeysize; n++)
      {
	sbidx1[i][n] =
	  (rol8 (key.d8[cnt], (i % 7) + 1) *
	   (ror8 (key.d8[cnt], (n % 7) + 1) + 1) + cnt) % 7;
	cnt++;
      }

  cnt = 0;
  x = 0;


  for (n = 0; n != subkeysize; n++)
    for (i = 0; i != subkeysize; i++)
      {

	x = ((x << n) ^ (x >> i)) + rol8 (key.d8[cnt], (cnt % 7) + 1) + cnt;
	sbidx2[i][n] = x % 7;
	cnt++;
      }


  pi = 0;

  for (n = 0; n != subkeysize; n++)
    {
      idx1 = idx_box1[n];
      idx2 = idx_box2[n];
      idx3 = idx_box3[n];
      idx4 = idx_box4[n];


      for (i = 0; i != subkeysize; i++, pi++)
	{

	  idx5 = sbidx1[n][i];
	  idx6 = sbidx2[n][i];

	  prefetch[pi].idx1 = idx1;
	  prefetch[pi].idx2 = idx2;
	  prefetch[pi].idx3 = idx3;
	  prefetch[pi].idx4 = idx4;
	  prefetch[pi].idx5 = idx5;
	  prefetch[pi].idx6 = idx6;
	  prefetch[pi].idx5_1 = idx5 + 1;
	  prefetch[pi].idx6_1 = idx6 + 1;


	  prefetch[pi].k1 = subkey1[idx1];
	  prefetch[pi].k2 = subkey1[idx1 + 1];
	  prefetch[pi].k3 = subkey2[idx2];
	  prefetch[pi].k4 = subkey2[idx2 + 1];

	  prefetch[pi].rot1 = rot_box[idx1 + idx2];
	  prefetch[pi].rot2 = rot_box[idx1 + idx2 + 1];

	  prefetch[pi].rot_box2 = rot_box2[idx3];
	  prefetch[pi].rot_box3 = rot_box3[idx4];



	  if (idx1 != (subkeysize - 2))
	    {
	      idx1++;
	    }
	  else
	    {
	      idx1 = 0;
	    }

	  if (idx2 != (subkeysize - 2))
	    {
	      idx2++;
	    }
	  else
	    {
	      idx2 = 0;
	    }


	  if (idx3 != (subkeysize - 2))
	    {
	      idx3++;
	    }
	  else
	    {
	      idx3 = 0;
	    }

	  if (idx4 != (subkeysize - 2))
	    {
	      idx4++;
	    }
	  else
	    {
	      idx4 = 0;
	    }


	}
    }



}


void __stdcall
setup (u32 * k)
{
  memmove (&key.d32[0], &k[0], keysizebytes);
  memmove (&opkey[0], &k[keysizebytes], 16);
  memmove (&subkey1[0], &key.d32[0], subkeysize * 4);
  memmove (&subkey2[0], &key.d32[subkeysize], subkeysize * 4);
  makebox ();
  makestruct ();

}

void __stdcall
killkey ()
{
  memset (&key.d8[0], 0x00, keysizebytes);
  memset (&prefetch, 0x00, sizeof (prefetch));
  memset (&oprec, 0x00, sizeof (oprec));
}



#define MIX16(X)\
\
rol32(X,16)\
\


__inline u32
MIX8 (const u32 v)
{
  u32 result;


  _asm
  {

    mov eax, v;
    bswap eax;
    mov result, eax;
  }

  return result;

}




#define transform(X, Y, N)\
\
switch (oprec[n][i][N]) {\
\
case add: X+=Y;\
break;\
	   \
case sub: X-=Y;\
break;\
	   \
case xor: X^=Y;\
break;\
\
case mixadd: X=MIX16(X)+Y;\
break;\
	   \
case mixsub: X=MIX16(X)-Y;\
break;\
	   \
case mixxor: X=MIX16(X)^Y;\
break;\
\
case mix8add: X=MIX8(X)+Y;\
break;\
	   \
case mix8sub: X=MIX8(X)-Y;\
break;\
	   \
case mix8xor: X=MIX8(X)^Y;\
break;\
\
}


#define restore(X, Y,  N)\
\
switch (oprec[n][i][N]) {\
\
case add: X-=Y;\
break;\
	   \
case sub: X+=Y;\
break;\
	   \
case xor: X^=Y;\
break;\
\
case mixadd: X=MIX16(X-Y);\
break;\
	   \
case mixsub: X=MIX16(X+Y);\
break;\
	   \
case mixxor: X=MIX16(X^Y);\
break;\
      \
case mix8add: X=MIX8(X-Y);\
break;\
	   \
case mix8sub: X=MIX8(X+Y);\
break;\
	   \
case mix8xor: X=MIX8(X^Y);\
break;\
}


void __stdcall
crypt (r32 * v)
{
  u32 i, n, v0, v1, j, m, s, t;
  u32 pi = 0;

  for (n = 0; n != subkeysize; n++)
    {
      for (i = 0; i != subkeysize; i++, pi++)
	{
	  s = v[0].d32 ^ v[1].d32 ^ v[2].d32 ^ v[3].d32;

	  m = s & 3;
	  t = v[0].d32;
	  v[0].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = (s >> 8) & 3;
	  t = v[1].d32;
	  v[1].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = (s >> 16) & 3;
	  t = v[2].d32;
	  v[2].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = (s >> 24) & 3;
	  t = v[3].d32;
	  v[3].d32 = v[m].d32;
	  v[m].d32 = t;


	  v[0].d8[0] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[0]];
	  v[0].d8[1] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[1]];
	  v[0].d8[2] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[2]];
	  v[0].d8[3] = SBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[3]];

	  v[1].d8[0] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[0]];
	  v[1].d8[1] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[1]];
	  v[1].d8[2] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[2]];
	  v[1].d8[3] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[3]];

	  v[2].d8[0] =
	    SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[0]];
	  v[2].d8[1] =
	    SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[1]];
	  v[2].d8[2] =
	    SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[2]];
	  v[2].d8[3] =
	    SBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[3]];

	  v[3].d8[0] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[0]];
	  v[3].d8[1] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[1]];
	  v[3].d8[2] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[2]];
	  v[3].d8[3] =
	    SBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[3]];


	  if (prefetch[pi].idx1 == idx1null)
	    {
	      transform (v[0].d32,
			 ((v[3].d32 + ((v[3].d32 << 6) ^ (v[3].d32 >> 10))) +
			  ror32 ((v[1].d32 +
				  ((v[1].d32 << 6) ^ (v[1].d32 >> 9))),
				 prefetch[pi].rot1) + prefetch[pi].k1), 0);
	    }
	  else
	    {
	      transform (v[0].d32,
			 ((v[3].d32 + ((v[3].d32 << 6) ^ (v[3].d32 >> 10))) +
			  ror32 ((v[1].d32 +
				  ((v[1].d32 << 6) ^ (v[1].d32 >> 9))) +
				 prefetch[pi].k1, prefetch[pi].rot1)), 1);
	    }

	  if (prefetch[pi].idx2 == idx2null)
	    {
	      transform (v[1].d32,
			 ((v[2].d32 + ((v[2].d32 << 6) ^ (v[2].d32 >> 10))) +
			  ror32 ((v[0].d32 +
				  ((v[0].d32 << 6) ^ (v[0].d32 >> 9))),
				 prefetch[pi].rot2) + prefetch[pi].k2), 2);
	    }
	  else
	    {
	      transform (v[1].d32,
			 ((v[2].d32 + ((v[2].d32 << 6) ^ (v[2].d32 >> 10))) +
			  ror32 ((v[0].d32 +
				  ((v[0].d32 << 6) ^ (v[0].d32 >> 9))) +
				 prefetch[pi].k2, prefetch[pi].rot2)), 3);
	    }


	  if (i == idx3null)
	    {
	      transform (v[2].d32,
			 rol32 (((v[0].d32 +
				  ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
				 prefetch[pi].k3) * 9, prefetch[pi].rot_box2),
			 4);
	      transform (v[3].d32,
			 rol32 (((v[1].d32 +
				  ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
				 prefetch[pi].k4) * 9, prefetch[pi].rot_box3),
			 5);
	    }
	  else
	    {
	      transform (v[2].d32,
			 rol32 (((v[0].d32 +
				  ((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
				 prefetch[pi].k3) * 9, prefetch[pi].rot_box2),
			 6);
	      transform (v[3].d32,
			 rol32 (((v[1].d32 +
				  ((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
				 prefetch[pi].k4) * 9, prefetch[pi].rot_box3),
			 7);
	    }


	  v0 = v[0].d32;
	  v1 = v[1].d32;
	  v[0].d32 = v[2].d32;
	  v[1].d32 = v[3].d32;
	  v[2].d32 = v0;
	  v[3].d32 = v1;

	}
    }
}

void __stdcall
decrypt (r32 * v)
{
  u32 v0, v1, m, s, t;
  u32 pi = rounds_total - 1;
  long n, i, j;

  for (n = subkeysize - 1; n != -1; n--)

    {

      for (i = subkeysize - 1; i != -1; i--, pi--)
	{

	  v0 = v[0].d32;
	  v1 = v[1].d32;
	  v[0].d32 = v[2].d32;
	  v[1].d32 = v[3].d32;
	  v[2].d32 = v0;
	  v[3].d32 = v1;

	  if (i == idx3null)
	    {
	      restore (v[3].d32,
		       rol32 (((v[1].d32 +
				((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
			       prefetch[pi].k4) * 9, prefetch[pi].rot_box3),
		       5);
	      restore (v[2].d32,
		       rol32 (((v[0].d32 +
				((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
			       prefetch[pi].k3) * 9, prefetch[pi].rot_box2),
		       4);
	    }
	  else
	    {
	      restore (v[3].d32,
		       rol32 (((v[1].d32 +
				((v[1].d32 << 4) ^ (v[1].d32 >> 5))) +
			       prefetch[pi].k4) * 9, prefetch[pi].rot_box3),
		       7);
	      restore (v[2].d32,
		       rol32 (((v[0].d32 +
				((v[0].d32 << 4) ^ (v[0].d32 >> 5))) +
			       prefetch[pi].k3) * 9, prefetch[pi].rot_box2),
		       6);
	    }



	  if (prefetch[pi].idx2 == idx2null)
	    {
	      restore (v[1].d32,
		       ((v[2].d32 + ((v[2].d32 << 6) ^ (v[2].d32 >> 10))) +
			ror32 ((v[0].d32 +
				((v[0].d32 << 6) ^ (v[0].d32 >> 9))),
			       prefetch[pi].rot2) + prefetch[pi].k2), 2);
	    }
	  else
	    {
	      restore (v[1].d32,
		       ((v[2].d32 + ((v[2].d32 << 6) ^ (v[2].d32 >> 10))) +
			ror32 ((v[0].d32 +
				((v[0].d32 << 6) ^ (v[0].d32 >> 9))) +
			       prefetch[pi].k2, prefetch[pi].rot2)), 3);
	    }



	  if (prefetch[pi].idx1 == idx1null)
	    {
	      restore (v[0].d32,
		       ((v[3].d32 + ((v[3].d32 << 6) ^ (v[3].d32 >> 10))) +
			ror32 ((v[1].d32 +
				((v[1].d32 << 6) ^ (v[1].d32 >> 9))),
			       prefetch[pi].rot1) + prefetch[pi].k1), 0);
	    }
	  else
	    {
	      restore (v[0].d32,
		       ((v[3].d32 + ((v[3].d32 << 6) ^ (v[3].d32 >> 10))) +
			ror32 ((v[1].d32 +
				((v[1].d32 << 6) ^ (v[1].d32 >> 9))) +
			       prefetch[pi].k1, prefetch[pi].rot1)), 1);
	    }


	  v[0].d8[0] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[0]];
	  v[0].d8[1] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[1]];
	  v[0].d8[2] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[2]];
	  v[0].d8[3] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6][v[0].d8[3]];

	  v[1].d8[0] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[0]];
	  v[1].d8[1] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[1]];
	  v[1].d8[2] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[2]];
	  v[1].d8[3] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6_1][v[1].d8[3]];

	  v[2].d8[0] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[0]];
	  v[2].d8[1] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[1]];
	  v[2].d8[2] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[2]];
	  v[2].d8[3] =
	    BackSBox[prefetch[pi].idx5][prefetch[pi].idx6_1][v[2].d8[3]];

	  v[3].d8[0] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[0]];
	  v[3].d8[1] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[1]];
	  v[3].d8[2] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[2]];
	  v[3].d8[3] =
	    BackSBox[prefetch[pi].idx5_1][prefetch[pi].idx6][v[3].d8[3]];

	  s = v[0].d32 ^ v[1].d32 ^ v[2].d32 ^ v[3].d32;

	  m = (s >> 24) & 3;
	  t = v[3].d32;
	  v[3].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = (s >> 16) & 3;
	  t = v[2].d32;
	  v[2].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = (s >> 8) & 3;
	  t = v[1].d32;
	  v[1].d32 = v[m].d32;
	  v[m].d32 = t;

	  m = s & 3;
	  t = v[0].d32;
	  v[0].d32 = v[m].d32;
	  v[m].d32 = t;

	}
    }
}



void
makestruct ()
{
  u32 x, y;
  keystream stream;
  u32 delta = 0x65DBAC36;

  u32 _oprec[subkeysize][subkeysize][8];

  for (x = 0; x < subkeysize; x++)
    for (y = 0; y < subkeysize; y++)
      {
	oprec[x][y][0] = add;
	oprec[x][y][1] = add;
	oprec[x][y][2] = add;
	oprec[x][y][3] = add;
	oprec[x][y][4] = xor;
	oprec[x][y][5] = xor;
	oprec[x][y][6] = add;
	oprec[x][y][7] = add;

      }

  y = delta;

  for (x = 0; x != 4; x++, y += delta)
    {

      stream.d32[x] = rol32 (opkey[x] + y, (x + 1) * 4);

    }

  for (x = 0; x < subkeysize; x++)
    for (y = 0; y < subkeysize; y++)

      {

	crypt ((r32 *) & stream.d32[0]);

	_oprec[x][y][0] = stream.d16[0] % 9;
	_oprec[x][y][1] = stream.d16[1] % 9;
	_oprec[x][y][2] = stream.d16[2] % 9;
	_oprec[x][y][3] = stream.d16[3] % 9;
	_oprec[x][y][4] = stream.d16[4] % 9;
	_oprec[x][y][5] = stream.d16[5] % 9;
	_oprec[x][y][6] = stream.d16[6] % 9;
	_oprec[x][y][7] = stream.d16[7] % 9;

      }

  for (x = 0; x < subkeysize; x++)
    for (y = 0; y < subkeysize; y++)
      {

	oprec[x][y][0] = _oprec[x][y][0];
	oprec[x][y][1] = _oprec[x][y][1];
	oprec[x][y][2] = _oprec[x][y][2];
	oprec[x][y][3] = _oprec[x][y][3];
	oprec[x][y][4] = _oprec[x][y][4];
	oprec[x][y][5] = _oprec[x][y][5];
	oprec[x][y][6] = _oprec[x][y][6];
	oprec[x][y][7] = _oprec[x][y][7];

      }

}
